import unittest
import fornax.opt as opt
import numpy as np


class TestProximity(unittest.TestCase):

    def setUp(self):
        self.h = 2
        self.alpha = .3

    def test_zero(self):
        proximities = opt._proximity(self.h, self.alpha, np.array([0]))
        self.assertListEqual([1.0], proximities.tolist())

    def test_one(self):
        proximities = opt._proximity(self.h, self.alpha, np.array([1]))
        self.assertListEqual([self.alpha], proximities.tolist())

    def test_pow(self):
        proximities = opt._proximity(self.h, self.alpha, np.array([2, 3, 4]))
        self.assertListEqual([self.alpha**2, 0.0, 0.0], proximities.tolist())

    def test_assert_h(self):
        self.assertRaises(ValueError, opt._proximity, -1, 0.3, np.array([0]))

    def test_assert_alpha_big(self):
        self.assertRaises(ValueError, opt._proximity, 2, 1.1, np.array([0]))

    def test_assert_alpha_small(self):
        self.assertRaises(ValueError, opt._proximity, 2, -.1, np.array([0]))


class TestDeltaPlus(unittest.TestCase):

    def test_greater(self):
        self.assertListEqual(
            opt._delta_plus(np.array([2, 4, 6]), np.array([1, 2, 3])).tolist(),
            [1, 2, 3]
        )

    def test_less(self):
        self.assertListEqual(
            opt._delta_plus(np.array([1, 2, 3]), np.array([2, 4, 6])).tolist(),
            [0, 0, 0]
        )


class TestNeighbourHoodMatchingCosts(unittest.TestCase):

    def setUp(self):
        self.records = [[1, 2, 3, 4, 5], [2, 4, 6, 7, 8]]
        self.costs = opt.NeighbourHoodMatchingCosts(self.records)

    def test_get_v(self):
        self.assertListEqual(list(self.costs.v), [1, 2])
        self.assertEqual(self.costs[0].v, 1)

    def test_get_u(self):
        self.assertListEqual(list(self.costs.u), [2, 4])
        self.assertEqual(self.costs[0].u, 2)

    def test_type(self):
        self.assertTrue(isinstance(
            self.costs[:1], opt.NeighbourHoodMatchingCosts))


class TestOpt(unittest.TestCase):
    """Reproduce the scenario set out in figure 4 of the paper"""

    def setUp(self):
        self.h = 2
        # use .5 because .5*.5=0.25 can be reproduced exactly in binary
        self.alpha = .5
        self.lmbda = .5
        self.records = ([
            (1, 1, 1, 1, 0, 0, 1), (1, 1, 1, 4, 0, 1, 1),
            (1, 1, 3, 3, 1, 1, 1), (1, 4, 1, 1, 0, 1, 1),
            (1, 4, 1, 4, 0, 0, 1), (1, 4, 2, 5, 1, 1, 1),
            (1, 4, 3, 6, 1, 1, 1), (1, 8, 1, 8, 0, 0, 1),
            (1, 8, 2, 9, 1, 1, 1), (1, 8, 3, 6, 1, 1, 1),
            (1, 8, 3, 12, 1, 1, 1), (2, 2, 2, 2, 0, 0, 1),
            (2, 2, 1, 1, 1, 1, 1), (2, 2, 4, None, 1, None, 1),
            (2, 5, 1, 4, 1, 1, 1), (2, 5, 2, 5, 0, 0, 1),
            (2, 5, 4, 7, 1, 1, 1), (2, 9, 1, 8, 1, 1, 1),
            (2, 9, 2, 9, 0, 0, 1), (2, 9, 4, 10, 1, 1, 1),
            (3, 3, 1, 1, 1, 1, 1), (3, 3, 3, 3, 0, 0, 1),
            (3, 6, 1, 4, 1, 1, 1), (3, 6, 1, 8, 1, 1, 1),
            (3, 6, 3, 6, 0, 0, 1), (3, 12, 1, 8, 1, 1, 1),
            (3, 12, 3, 12, 0, 0, 1), (3, 13, 1, None, 1, None, 1),
            (3, 13, 3, 13, 0, 0, 1), (4, 7, 2, 5, 1, 1, 1),
            (4, 7, 4, 7, 0, 0, 1), (4, 7, 4, 10, 0, 1, 1),
            (4, 10, 2, 9, 1, 1, 1), (4, 10, 4, 7, 0, 1, 1),
            (4, 10, 4, 10, 0, 0, 1), (4, 10, 5, 11, 1, 1, 1),
            (5, 11, 4, 10, 1, 1, 1), (5, 11, 5, 11, 0, 0, 1)
        ])

    def test_neighbourhood_matching_costs(self):

        vals = [
            (1. - self.lmbda) * (1. - self.alpha),
            (1. - self.lmbda) * self.alpha
        ]
        target = [
            (1, 1, 1, 1, 0), (1, 1, 1, 4, vals[0]), (1, 1, 3, 3, 0),
            (1, 4, 1, 1, vals[0]), (1, 4, 1, 4, 0), (1, 4, 2, 5, 0),
            (1, 4, 3, 6, 0), (1, 8, 1, 8, 0), (1, 8, 2, 9, 0),
            (1, 8, 3, 6, 0), (1, 8, 3, 12, 0), (2, 2, 1, 1, 0),
            (2, 2, 2, 2, 0), (2, 2, 4, -1, vals[1]), (2, 5, 1, 4, 0),
            (2, 5, 2, 5, 0), (2, 5, 4, 7, 0), (2, 9, 1, 8, 0),
            (2, 9, 2, 9, 0), (2, 9, 4, 10, 0), (3, 3, 1, 1, 0),
            (3, 3, 3, 3, 0), (3, 6, 1, 4, 0), (3, 6, 1, 8, 0),
            (3, 6, 3, 6, 0), (3, 12, 1, 8, 0), (3, 12, 3, 12, 0),
            (3, 13, 1, -1, vals[1]), (3, 13, 3, 13, 0), (4, 7, 2, 5, 0),
            (4, 7, 4, 7, 0), (4, 7, 4, 10, vals[0]),  (4, 10, 2, 9, 0),
            (4, 10, 4, 7, vals[0]), (4, 10, 4, 10, 0), (4, 10, 5, 11, 0),
            (5, 11, 4, 10, 0), (5, 11, 5, 11, 0)
        ]
        result, _, _ = opt._get_matching_costs(
            self.records, 1, self.lmbda, self.alpha)
        self.assertListEqual(result.tolist(), target)

    def test_beta_1(self):

        target = {1: 2.0, 2: 2.0, 3: 1.5, 4: 2.0, 5: 1.5}
        _, _, beta = opt._get_matching_costs(
            self.records, 1, self.lmbda, self.alpha)
        result = {k: v for k, v in zip(range(1, 6), beta(range(1, 6)))}
        self.assertDictEqual(result, target)

    def test_optimal_matches(self):
        inference_costs, subgraphs, _, sz, target_edges = opt.solve(
            self.records, hopping_distance=1)
        perfect = []
        for sub_graph in subgraphs:
            cost = sz - \
                len(sub_graph) + sum(inference_costs[item]
                                     for item in sub_graph) / len(sub_graph)
            if cost == 0:
                perfect.append(sub_graph)

        self.assertSequenceEqual(
            perfect[0],
            [(1, 8), (2, 9), (3, 6), (4, 10), (5, 11)]
        )

        self.assertSequenceEqual(
            perfect[1],
            [(1, 8), (2, 9), (3, 12), (4, 10), (5, 11)]
        )



import unittest
import json
import fornax.api
import fornax.model
from test_base import TestCaseDB
from sqlalchemy.orm.session import Session
from unittest import TestCase


class DummyException(Exception):
    pass


class TestConnection(TestCaseDB):

    def test_rollback(self):
        """ Test than the connection rolls back
        transactions if theres an exception
        """
        dburl = 'sqlite://'
        try:
            # raise an exception afer creatiion
            # then raise an exception
            with fornax.Connection(dburl) as conn:
                graph = fornax.GraphHandle.create(conn)
                names = ['adam', 'ben', 'chris']
                graph.add_nodes(name=names)
                raise DummyException
        except DummyException:
            pass
        finally:
            # everything should be gone
            with fornax.Connection(dburl) as conn:
                n_nodes = conn.session.query(fornax.model.Node).count()
                n_graphs = conn.session.query(fornax.model.Graph).count()
            self.assertEqual(n_nodes, 0)
            self.assertEqual(n_graphs, 0)


class TestGraph(TestCaseDB):

    def run(self, result=None):
        with fornax.Connection('sqlite://') as conn:
            self.conn = conn
            self.conn.make_session = lambda: Session(self._connection)
            super().run(result)

    def test_init_raises(self):
        """ raise an ValueError if a hadle to a graph
        is constructed that does not exist
        """
        self.assertRaises(ValueError, fornax.GraphHandle, self.conn, 0)
        self.assertRaises(ValueError, fornax.GraphHandle.read, self.conn, 0)

    def test_create(self):
        """first graph has id zero
        """
        graph = fornax.GraphHandle.create(self.conn)
        self.assertEqual(graph.graph_id, 0)

    def test_create_two(self):
        """auto increment graph id
        """
        _ = fornax.GraphHandle.create(self.conn)
        second = fornax.GraphHandle.create(self.conn)
        self.assertEqual(second.graph_id, 1)

    def test_read(self):
        """get a graph handle using graph id
        """
        graph = fornax.GraphHandle.create(self.conn)
        graph_id = graph.graph_id
        same_graph = fornax.GraphHandle.read(self.conn, graph_id)
        self.assertEqual(same_graph.graph_id, graph_id)

    def test_delete(self):
        """getting a deleted graph should raise a value error
        """
        graph = fornax.GraphHandle.create(self.conn)
        graph.delete()
        self.assertRaises(ValueError, fornax.GraphHandle.read, self.conn, 0)

        graph = fornax.GraphHandle.create(self.conn)
        graph.add_nodes(id_src=[0, 1, 2])
        graph.add_edges([0, 2], [1, 1])
        graph.delete()

    def test_add_nodes(self):
        """meta data is stored on a node
        """
        graph = fornax.GraphHandle.create(self.conn)
        names = ['adam', 'ben', 'chris']
        graph.add_nodes(name=names)
        nodes = self.conn.session.query(fornax.model.Node).filter(
            fornax.model.Node.graph_id == 0).all()
        nodes = sorted(nodes, key=lambda node: node.node_id)
        self.assertListEqual(
            names, [json.loads(node.meta)['name'] for node in nodes])

    def test_add_nodes_more_meta(self):
        """multiple metadata is stored on a node
        """
        graph = fornax.GraphHandle.create(self.conn)
        names = ['adam', 'ben', 'chris']
        ages = [9, 10, 11]
        graph.add_nodes(name=names, age=ages)
        nodes = self.conn.session.query(fornax.model.Node).filter(
            fornax.model.Node.graph_id == 0).all()
        nodes = sorted(nodes, key=lambda node: node.node_id)
        self.assertListEqual(
            names, [json.loads(node.meta)['name'] for node in nodes])
        self.assertListEqual(
            ages, [json.loads(node.meta)['age'] for node in nodes])

    def test_missing_attribute(self):
        """Null values for metadata must be explicit
        """
        graph = fornax.GraphHandle.create(self.conn)
        names = ['adam', 'ben', 'chris']
        ages = [9, 10]
        self.assertRaises(TypeError, graph.add_nodes, name=names, age=ages)

    def test_assign_id(self):
        """assigning node id is forbidden
        """
        graph = fornax.GraphHandle.create(self.conn)
        ids = range(3)
        self.assertRaises(ValueError, graph.add_nodes, id=ids)

    def test_no_metadata(self):
        """Nodes must have some metadata associated with them
        """
        graph = fornax.GraphHandle.create(self.conn)
        self.assertRaises(ValueError, graph.add_nodes)

    def test_add_edges(self):
        """store metadata on edges
        """
        graph = fornax.GraphHandle.create(self.conn)
        names = ['adam', 'ben', 'chris']
        ages = [9, 10, 11]
        graph.add_nodes(name=names, age=ages)
        relationships = ['is_friend', 'is_foe']
        graph.add_edges([0, 0], [1, 2], relationship=relationships)
        edges = self.conn.session.query(
            fornax.model.Edge
        ).filter(
            fornax.model.Edge.graph_id == graph.graph_id
        ).filter(
            fornax.model.Edge.start < fornax.model.Edge.end
        ).all()
        edges = sorted(edges, key=lambda edge: (edge.start, edge.end))
        self.assertListEqual(relationships, [json.loads(
            edge.meta)['relationship'] for edge in edges])

    def test_add_edges_more_meta(self):
        """store multiple metadata on edges
        """
        graph = fornax.GraphHandle.create(self.conn)
        names = ['adam', 'ben', 'chris']
        ages = [9, 10, 11]
        graph.add_nodes(name=names, age=ages)
        relationships = ['is_friend', 'is_foe']
        types = [0, 1]
        graph.add_edges(
            [0, 0], [1, 2], relationship=relationships, type_=types)
        edges = self.conn.session.query(
            fornax.model.Edge
        ).filter(
            fornax.model.Edge.graph_id == graph.graph_id
        ).filter(
            fornax.model.Edge.start < fornax.model.Edge.end
        ).all()
        edges = sorted(edges, key=lambda edge: (edge.start, edge.end))
        self.assertListEqual(relationships, [json.loads(
            edge.meta)['relationship'] for edge in edges])
        self.assertListEqual(
            types, [json.loads(edge.meta)['type_'] for edge in edges])

    def test_simple_graph(self):
        """Test for a simple graph.
        A simple graph is a graph with no loops.
        A loop is an edge that connects a vertex to itself
        """
        graph = fornax.GraphHandle.create(self.conn)
        names = ['adam', 'ben', 'chris']
        ages = [9, 10, 11]
        graph.add_nodes(name=names, age=ages)
        self.assertRaises(fornax.api.InvalidEdgeError, graph.add_edges, [
                          1, 0], [1, 2], relationship=['is_friend', 'is_foe'])

    def test_add_nodes_id_src(self):
        graph = fornax.GraphHandle.create(self.conn)
        graph.add_nodes(id_src=['a', 'b', 'c', 'd'])
        graph.add_edges(['a', 'b'], ['b', 'c'])
        nodes = self.conn.session.query(fornax.model.Node).all()
        self.assertEqual(
            [n.node_id for n in nodes],
            [self.conn._hash(item) for item in ('a', 'b', 'c', 'd')]
        )

    def test_add_nodes_id_src_meta(self):
        graph = fornax.GraphHandle.create(self.conn)
        graph.add_nodes(id_src=['a', 'b', 'c', 'd'])
        graph.add_edges(['a', 'b'], ['b', 'c'])
        nodes = self.conn.session.query(fornax.model.Node).all()
        self.assertEqual(
            [json.loads(n.meta)['id_src'] for n in nodes],
            ['a', 'b', 'c', 'd']
        )

    def test_add_edges_id_src(self):
        graph = fornax.GraphHandle.create(self.conn)
        graph.add_nodes(id_src=['a', 'b', 'c', 'd'])
        graph.add_edges(['a', 'b'], ['b', 'c'])
        edges = self.conn.session.query(
            fornax.model.Edge
        ).filter(
            fornax.model.Edge.start < fornax.model.Edge.end
        ).all()
        self.assertEqual(
            sorted([e.start, e.end] for e in edges),
            sorted(
                sorted([self.conn._hash(start), self.conn._hash(end)])
                for start, end in [('a', 'b'), ('b', 'c')]
            )
        )          


class TestQuery(TestCaseDB):

    def run(self, result=None):
        with fornax.Connection('sqlite://') as conn:
            self.conn = conn
            self.conn.make_session = lambda: Session(self._connection)
            super().run(result)

    def test_init_query_raises(self):
        self.assertRaises(ValueError, fornax.QueryHandle, self.conn, 0)

    def test_init_read_raises(self):
        self.assertRaises(ValueError, fornax.QueryHandle.read, self.conn, 0)

    def test_create(self):
        query_graphs = [fornax.GraphHandle.create(self.conn) for _ in range(3)]
        target_graphs = [
            fornax.GraphHandle.create(self.conn) for _ in range(3)
        ]
        queries = [fornax.QueryHandle.create(
            self.conn, q, t) for q, t in zip(query_graphs, target_graphs)]
        self.assertEqual([q.query_id for q in queries], [0, 1, 2])

    def test_create_query_target(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        query_db = self.conn.session.query(fornax.model.Query).filter(
            fornax.model.Query.query_id == query.query_id).first()
        self.assertEqual(query_db.start_graph_id, query_graph.graph_id)
        self.assertEqual(query_db.end_graph_id, target_graph.graph_id)

    def test_read(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        q1 = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        q2 = fornax.QueryHandle.read(self.conn, q1.query_id)
        self.assertEqual(q1.query_id, q2.query_id)

    def test_delete(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        query_graph.add_nodes(id_src=[0, 1])
        query_graph.add_edges([0], [1])
        target_graph = fornax.GraphHandle.create(self.conn)
        target_graph.add_nodes(id_src=[1, 2])
        target_graph.add_edges([2], [1])
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        query.add_matches([0, 1], [2, 1], [1, 1])
        query_id = query.query_id
        query.delete()
        query_exists = self.conn.session.query(fornax.model.Query).filter(
            fornax.model.Query.query_id == query_id).scalar()
        matches_exists = self.conn.session.query(fornax.model.Match).filter(
            fornax.model.Match.query_id == query_id).scalar()
        self.assertFalse(query_exists)
        self.assertFalse(matches_exists)

    def test_get_query_graph(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        self.assertEqual(query.query_graph(), query_graph)

    def test_get_target_graph(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        self.assertEqual(query.target_graph(), target_graph)

    def test_query_nodes(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        q_uids = [0, 1, 2]
        t_uids = [3, 4, 5]
        query_graph.add_nodes(uid=q_uids)
        target_graph.add_nodes(uid=t_uids)
        query_nodes = query._query_nodes()
        self.assertListEqual(
            [fornax.api.Node(i, 'query', {'uid': uid})
             for i, uid in enumerate(q_uids)],
            query_nodes
        )

    def test_query_edges(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        q_uids = [0, 1, 2]
        t_uids = [3, 4, 5]
        query_graph.add_nodes(uid=q_uids)
        query_graph.add_edges([0, 0], [1, 2], my_id=['a', 'b'])
        target_graph.add_nodes(uid=t_uids)
        query_edges = query._query_edges()
        self.assertListEqual(
            [fornax.api.Edge(0, 1, 'query', {'my_id': 'a'}), fornax.api.Edge(
                0, 2, 'query', {'my_id': 'b'})],
            query_edges
        )

    def test_target_nodes(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        q_uids = [0, 1, 2]
        t_uids = [3, 4, 5]
        query_graph.add_nodes(id_src=q_uids)
        target_graph.add_nodes(id_src=t_uids)
        #  no target nodes will appear if there are no matches
        query.add_matches(q_uids, t_uids, [1, 1, 1])
        target_nodes = query._target_nodes()
        self.assertListEqual(
            [
                fornax.api.Node(uid, 'target', {'id_src': uid})
                for uid in t_uids
            ],
            target_nodes
        )

    def test_undirected_edges(self):
        """Each edge needs to be stored in both directions
        """
        graph = fornax.GraphHandle.create(self.conn)
        graph.add_nodes(myid=[1, 2, 3])
        graph.add_edges([0], [1])
        src = [
            (e.start, e.end)
            for e in self.conn.session.query(fornax.model.Edge).all()
        ]
        self.assertListEqual(sorted(src), [(0, 1), (1, 0)])

    def test_target_edges(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        uids = [0, 1]
        query_graph.add_nodes(uid=range(3))
        target_graph.add_nodes(uid=range(3))
        target_graph.add_edges([0, 1], [1, 2])
        sources = [0, 0]
        targets = [1, 2]
        weights = [1, 1]
        query.add_matches(sources, targets, weights, my_id=uids)
        _, _, _, _, target_edges_arr = query._optimise(2, 10, None)
        target_edges = query._target_edges(
            query._target_nodes(), target_edges_arr)
        self.assertListEqual(
            [
                #  no matches end on node 0
                #  fornax.api.Edge(0, 1, 'target', dict()),
                fornax.api.Edge(1, 2, 'target', dict())
            ],
            target_edges
        )

    def test_add_matches(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        uids = [0, 1]
        query_graph.add_nodes(uid=range(3))
        target_graph.add_nodes(uid=range(3))
        sources = [0, 0]
        targets = [1, 2]
        weights = [1, 1]
        query.add_matches(sources, targets, weights, my_id=uids)
        matches = self.conn.session.query(fornax.model.Match).filter(
            fornax.model.Match.query_id == query.query_id
        ).order_by(fornax.model.Match.end.asc())
        self.assertEqual(sources, [m.start for m in matches])
        self.assertEqual(targets, [m.end for m in matches])
        self.assertEqual(weights, [m.weight for m in matches])
        self.assertEqual(uids, [json.loads(m.meta)['my_id'] for m in matches])

    def test_execute_raises(self):
        query_graph = fornax.GraphHandle.create(self.conn)
        target_graph = fornax.GraphHandle.create(self.conn)
        query = fornax.QueryHandle.create(self.conn, query_graph, target_graph)
        self.assertRaises(ValueError, query.execute)


class TestNode(TestCase):

    def setUp(self):
        self.node = fornax.api.Node(0, 'query', {'a': 1})

    def test_id(self):
        self.assertEqual(self.node.id, 0)

    def test_meta(self):
        self.assertEqual(self.node.meta, {'a': 1})

    def test_eq(self):
        self.assertEqual(self.node, fornax.api.Node(0, 'query', {'a': 1}))
        self.assertNotEqual(self.node, fornax.api.Node(1, 'query', {'a': 1}))
        self.assertNotEqual(self.node, fornax.api.Node(0, 'query', {'a': 0}))
        self.assertNotEqual(self.node, fornax.api.Node(1, 'query', {'a': 0}))
        self.assertNotEqual(self.node, fornax.api.Node(0, 'target', {'a': 1}))

    def test_node_raises(self):
        self.assertRaises(ValueError, fornax.api.Node, 0, 'a', {})


class TestEdge(TestCase):

    def setUp(self):
        self.edge = fornax.api.Edge(0, 1, 'query', {'a': 1})

    def test_start(self):
        self.assertEqual(self.edge.start, 0)

    def test_end(self):
        self.assertEqual(self.edge.end, 1)

    def test_meta(self):
        self.assertEqual(self.edge.meta, {'a': 1})

    def test_eq(self):
        self.assertEqual(self.edge, fornax.api.Edge(0, 1, 'query', {'a': 1}))
        self.assertNotEqual(
            self.edge, fornax.api.Edge(1, 1, 'query', {'a': 1}))
        self.assertNotEqual(
            self.edge, fornax.api.Edge(0, 0, 'query', {'a': 1}))
        self.assertNotEqual(
            self.edge, fornax.api.Edge(0, 1, 'query', {'a': 2}))
        self.assertNotEqual(
            self.edge, fornax.api.Edge(0, 1, 'target', {'a': 1}))

    def test_edge_raises(self):
        self.assertRaises(ValueError, fornax.api.Edge, 0, 1, 'a', {})


class TestExample(TestCaseDB):

    @classmethod
    def setUp(self):
        """trick fornax into using the test database setup
        """
        super().setUp(self)
        self.maxsize = fornax.Connection.SQLITE_MAX_SIZE
        with fornax.Connection('sqlite://') as conn:
            conn.make_session = lambda: Session(self._connection)
            query_graph = fornax.GraphHandle.create(conn)
            query_graph.add_nodes(my_id=range(1, 6))
            starts, ends = zip(*[(1, 3), (1, 2), (2, 4), (4, 5)])
            query_graph.add_edges(
                [i - 1 for i in starts],
                [i - 1 for i in ends]
            )

            target_graph = fornax.GraphHandle.create(conn)
            target_graph.add_nodes(my_id=range(1, 14))
            starts, ends = zip(*[
                (1, 2), (1, 3), (1, 4),
                (3, 7), (4, 5), (4, 6),
                (5, 7), (6, 8), (7, 10),
                (8, 9), (8, 12), (9, 10),
                (10, 11), (11, 12), (11, 13),
            ])
            target_graph.add_edges(
                [s - 1 for s in starts],
                [e - 1 for e in ends]
            )

            query = fornax.QueryHandle.create(
                conn,
                query_graph,
                target_graph
            )
            starts, ends, weights = zip(*[
                (1, 1, 1), (1, 4, 1), (1, 8, 1),
                (2, 2, 1), (2, 5, 1), (2, 9, 1),
                (3, 3, 1), (3, 6, 1), (3, 12, 1), (3, 13, 1),
                (4, 7, 1), (4, 10, 1),
                (5, 11, 1)
            ])

            query.add_matches(
                [s - 1 for s in starts],
                [e - 1 for e in ends],
                weights
            )

            self.payload = query.execute(n=2)

    def test_iters(self):
        self.assertEqual(self.payload['max_iters'], 10)

    def test_hopping_distance(self):
        self.assertEqual(self.payload['hopping_distance'], 2)

    def test_first_graph_cost(self):
        graph = self.payload['graphs'][0]
        self.assertEqual(graph['cost'], 0)

    def test_first_graph_nodes(self):
        graph = self.payload['graphs'][1]
        nodes = [
            {"id": 0, "type": "query", "my_id": 1},
            {"id": 1, "type": "query", "my_id": 2},
            {"id": 2, "type": "query", "my_id": 3},
            {"id": 3, "type": "query", "my_id": 4},
            {"id": 4, "type": "query", "my_id": 5},
            {"id": 7, "type": "target", "my_id": 8},
            {"id": 8, "type": "target", "my_id": 9},
            {"id": 9, "type": "target", "my_id": 10},
            {"id": 10, "type": "target", "my_id": 11},
            {"id": 11, "type": "target", "my_id": 12}
        ]
        for node in nodes:
            node['id'] = fornax.api._hash(
                (node['id'], node['type']),
                self.maxsize
            )
        self.assertListEqual(
            graph['nodes'],
            nodes
        )

    def test_first_graph_links(self):

        graph = self.payload['graphs'][1]

        matches = [
            {"source": 0, "target": 7, "type": "match", "weight": 1.0},
            {"source": 1, "target": 8, "type": "match", "weight": 1.0},
            {"source": 2, "target": 11, "type": "match", "weight": 1.0},
            {"source": 3, "target": 9, "type": "match", "weight": 1.0},
            {"source": 4, "target": 10, "type": "match", "weight": 1.0},
            {"source": 0, "target": 1, "type": "query", "weight": 1.0},
            {"source": 0, "target": 2, "type": "query", "weight": 1.0},
            {"source": 1, "target": 3, "type": "query", "weight": 1.0},
            {"source": 3, "target": 4, "type": "query", "weight": 1.0},
            {"source": 7, "target": 8, "type": "target", "weight": 1.0},
            {"source": 7, "target": 11, "type": "target", "weight": 1.0},
            {"source": 8, "target": 9, "type": "target", "weight": 1.0},
            {"source": 9, "target": 10, "type": "target", "weight": 1.0},
            {"source": 10, "target": 11, "type": "target", "weight": 1.0},
        ]

        for match in matches:

            if match['type'] == 'query' or match['type'] == 'target':

                match['source'] = fornax.api._hash(
                    (match['source'], match['type']),
                    self.maxsize
                )

                match['target'] = fornax.api._hash(
                    (match['target'], match['type']),
                    self.maxsize
                )

            else:

                match['source'] = fornax.api._hash(
                    (match['source'], 'query'),
                    self.maxsize
                )

                match['target'] = fornax.api._hash(
                    (match['target'], 'target'),
                    self.maxsize
                )

        self.assertListEqual(graph['links'], matches)

    def test_second_graph_cost(self):
        graph = self.payload['graphs'][1]
        self.assertEqual(graph['cost'], 0)

    def test_second_graph_nodes(self):
        graph = self.payload['graphs'][0]
        nodes = [
            {"id": 0, "type": "query", "my_id": 1},
            {"id": 1, "type": "query", "my_id": 2},
            {"id": 2, "type": "query", "my_id": 3},
            {"id": 3, "type": "query", "my_id": 4},
            {"id": 4, "type": "query", "my_id": 5},
            {"id": 5, "type": "target", "my_id": 6},
            {"id": 7, "type": "target", "my_id": 8},
            {"id": 8, "type": "target", "my_id": 9},
            {"id": 9, "type": "target", "my_id": 10},
            {"id": 10, "type": "target", "my_id": 11},
        ]
        for node in nodes:
            node['id'] = fornax.api._hash(
                (node['id'], node['type']),
                self.maxsize
            )

        self.assertListEqual(
            graph['nodes'],
            nodes
        )

    def test_second_graph_links(self):
        graph = self.payload['graphs'][0]
        matches = [
            {"source": 0, "target": 7, "type": "match", "weight": 1.0},
            {"source": 1, "target": 8, "type": "match", "weight": 1.0},
            {"source": 2, "target": 5, "type": "match", "weight": 1.0},
            {"source": 3, "target": 9, "type": "match", "weight": 1.0},
            {"source": 4, "target": 10, "type": "match", "weight": 1.0},
            {"source": 0, "target": 1, "type": "query", "weight": 1.0},
            {"source": 0, "target": 2, "type": "query", "weight": 1.0},
            {"source": 1, "target": 3, "type": "query", "weight": 1.0},
            {"source": 3, "target": 4, "type": "query", "weight": 1.0},
            {"source": 5, "target": 7, "type": "target", "weight": 1.0},
            {"source": 7, "target": 8, "type": "target", "weight": 1.0},
            {"source": 8, "target": 9, "type": "target", "weight": 1.0},
            {"source": 9, "target": 10, "type": "target", "weight": 1.0},
        ]

        for match in matches:

            if match['type'] == 'query' or match['type'] == 'target':

                match['source'] = fornax.api._hash(
                    (match['source'], match['type']),
                    self.maxsize
                )

                match['target'] = fornax.api._hash(
                    (match['target'], match['type']),
                    self.maxsize
                )

            elif match['type'] == 'match':

                match['source'] = fornax.api._hash(
                    (match['source'], 'query'),
                    self.maxsize
                )

                match['target'] = fornax.api._hash(
                    (match['target'], 'target'),
                    self.maxsize
                )

        self.assertListEqual(graph['links'], matches)



import sqlalchemy
import unittest
import fornax.model as model
from test_base import TestCaseDB


class TestNode(TestCaseDB):

    def test_node_round_trip(self):
        """ node round trip """
        new_graph = model.Graph(graph_id=0)
        self.session.add(new_graph)
        self.session.commit()
        new_node = model.Node(node_id=0, graph_id=0)
        self.session.add(new_node)
        self.session.commit()

        row = self.session.query(model.Node).first()
        self.assertIsNotNone(row)


class TestEdge(TestCaseDB):

    def setUp(self):
        super().setUp()
        new_graphs = [model.Graph(graph_id=0), model.Graph(graph_id=1)]
        self.session.add_all(new_graphs)
        self.session.commit()
        new_nodes = [model.Node(node_id=id_, graph_id=0) for id_ in range(2)]
        new_nodes += [model.Node(node_id=id_, graph_id=1) for id_ in range(2)]
        self.session.add_all(new_nodes)
        self.session.commit()

        new_edges = [
            model.Edge(start=0, end=1, graph_id=0),
            model.Edge(start=1, end=0, graph_id=1),
        ]
        self.session.add_all(new_edges)
        self.session.commit()

    def test_query_edge_round_trip(self):
        """ edge round trip """
        row = self.session.query(model.Edge).first()
        self.assertIsNotNone(row)
        self.assertEqual(0, row.start)

    def test_edge_join_start(self):
        """ find a node by joining on the start of an edge """
        query = self.session.query(model.Node)
        query = query.join(
            model.Edge,
            sqlalchemy.and_(
                model.Node.node_id == model.Edge.start,
                model.Node.graph_id == model.Edge.graph_id
            )
        )
        row = query.first()
        self.assertIsNotNone(row)
        self.assertEqual(row.node_id, 0)

    def test_edge_join_end(self):
        """ find a node by joining on the end of an edge """
        query = self.session.query(model.Node)
        query = query.join(
            model.Edge,
            sqlalchemy.and_(
                model.Node.node_id == model.Edge.end,
                model.Node.graph_id == model.Edge.graph_id
            )
        )
        row = query.first()
        self.assertIsNotNone(row)
        self.assertEqual(row.node_id, 1)


class TestNeighbours(TestCaseDB):

    def setUp(self):
        super().setUp()
        new_graph = model.Graph(graph_id=0)
        self.session.add(new_graph)
        self.session.commit()
        new_nodes = [model.Node(node_id=id_, graph_id=0) for id_ in range(4)]
        self.session.add_all(new_nodes)
        self.session.commit()

        new_edges = [
            model.Edge(start=0, end=1, graph_id=0),
            model.Edge(start=0, end=2, graph_id=0),
            model.Edge(start=2, end=3, graph_id=0),
        ]

        self.session.add_all(new_edges)
        self.session.commit()

    def test_neighbours(self):
        query = self.session.query(model.Node)
        query = query.filter(model.Node.node_id == 0)
        node = query.first()
        self.assertListEqual(
            [n.node_id for n in node.neighbours()],
            [1, 2]
        )

    def test_next_neighbours(self):
        query = self.session.query(model.Node)
        query = query.filter(model.Node.node_id == 0)
        node = query.first()
        self.assertListEqual(
            [n2.node_id for n1 in node.neighbours() for n2 in n1.neighbours()],
            [3]
        )


class TestMatch(TestCaseDB):

    def setUp(self):
        super().setUp()
        new_graphs = [model.Graph(graph_id=0), model.Graph(graph_id=1)]
        self.session.add_all(new_graphs)
        self.session.commit()
        new_query = model.Query(query_id=0, start_graph_id=0, end_graph_id=1)
        self.session.add(new_query)
        self.session.commit()

        new_nodes = [
            model.Node(node_id=0, graph_id=0),
            model.Node(node_id=0, graph_id=1),
            model.Node(node_id=1, graph_id=0),
            model.Node(node_id=1, graph_id=1)
        ]
        self.session.add_all(new_nodes)
        self.session.commit()

        new_edges = [
            model.Match(start=0, end=0, weight=1, start_graph_id=0,
                        end_graph_id=1, query_id=0),
            model.Match(start=1, end=0, weight=1, start_graph_id=0,
                        end_graph_id=1, query_id=0),
            model.Match(start=1, end=1, weight=1, start_graph_id=0,
                        end_graph_id=1, query_id=0)
        ]
        self.session.add_all(new_edges)
        self.session.commit()

    def test_select_by_match_left(self):
        query = self.session.query(model.Node)
        query = query.filter(
            model.Node.start_matches.any(
                model.Match.start == 0
            )
        )
        nodes = query.all()
        self.assertLessEqual(
            [node.node_id for node in nodes],
            [0, 1]
        )

    def test_test_min_check(self):
        self.session.add(model.Match(start=0, end=0, weight=1.1,
                                     start_graph_id=0, end_graph_id=0))
        self.assertRaises(
            sqlalchemy.exc.IntegrityError,
            self.session.commit
        )

    def test_test_max_check(self):
        self.session.add(model.Match(start=0, end=0, weight=0,
                                     start_graph_id=0, end_graph_id=0))
        self.assertRaises(
            sqlalchemy.exc.IntegrityError,
            self.session.commit
        )


if __name__ == '__main__':
    unittest.main()



import unittest
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm.session import Session
import fornax.model as model


Base = model.Base


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


class TestCaseDB(unittest.TestCase):
    """ A base test for setting up and tearing down the databse """

    @classmethod
    def setUpClass(cls):
        """ Create the engine, create one connection
        and start a transaction """
        engine = create_engine(
            'sqlite://',
            echo=False
        )

        connection = engine.connect()
        cls._engine = engine
        cls._connection = connection
        cls.__transaction = connection.begin()
        Base.metadata.create_all(connection)

    @classmethod
    def tearDownClass(cls):
        """ tear down the top level transaction """
        cls.__transaction.rollback()
        cls._connection.close()
        cls._engine.dispose()

    def setUp(self):
        """ create a new session and a nested transaction """
        self._transaction = self._connection.begin_nested()
        self.session = Session(self._connection)

    def tearDown(self):
        """ rollback the nested transaction """
        self._transaction.rollback()
        self.session.close()



import numpy as np
from itertools import starmap
from typing import List


def _proximity(h: float, alpha: float, distances: np.ndarray) -> np.ndarray:
    """Calculates the proximity factor P for an array of distances.
    Implements equation 1 in the paper

    Arguments:
        h {float} -- max hopping distance
        alpha {float} -- propagation factor
        distances {np.array} -- array of hopping distances

    Raises:
        ValueError -- if hopping distance is less than zero
        ValueError -- if propagation factor is not between zero and one

    Returns:
        np.array -- an array of proximiy values
    """

    if h < 0:
        raise ValueError('hopping distance h cannot be negative')
    if not 0 < alpha <= 1:
        raise ValueError('propagation factor alpha must be between 0 and 1')
    return np.multiply(
        np.less_equal(distances, h),
        np.power(alpha, distances)
    )


def _delta_plus(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Comparator function. Equation 3 in the paper.

    Arguments:
        x {np.array} -- an array of floats
        y {np.array} -- an array of floats

    Returns
        np.array -- an array of floats
    """

    return np.multiply(
        np.greater(x, y),
        np.subtract(x, y)
    )


def group_by(columns, arr):
    """Split an array into n slices where 'columns'
    are all equal within each slice

    Arguments:
        columns {List[str]} -- a list of column names
        arr {np.array} -- a numpy structured array

    Returns
        keys: np.array -- the column values uniquly identifying each group
        groups: List[np.array] -- a list of numpy arrays
    """

    if not len(columns):
        raise ValueError("group_by requires a non empty list of column names")

    _, counts = np.unique(arr[columns], return_counts=True)
    indices = np.insert(np.cumsum(counts), 0, 0)
    split = np.split(arr, indices)
    filtered = list(filter(lambda x: len(x), split))
    keys = map(lambda x: x[columns][0].tolist(), filtered)
    stacked = np.vstack(v for v in keys)
    return stacked, filtered


def group_by_first(columns, arr):
    """
    Split an array into n slices where 'columns'
    all compare equal within each slide
    Take the first row of each slice
    Combine each of the rows into a single array through concatination

    Arguments:
        columns {[str]} -- a list of column names
        arr {[type]} -- a numpy structured array

    Returns:
        np.array - new concatinated array
    """
    _, counts = np.unique(arr[columns], return_counts=True)
    indices = np.cumsum(np.insert(counts, 0, 0))[:-1]
    return arr[indices]


class Base(np.recarray):

    """A Base class for subclassing numpy record arrays

    Returns:
        np.recarray -- A subclass of np.recarray
    """

    columns = []
    types = []

    def __new__(cls, *args, **kwargs):
        dtype = list(zip(cls.columns, cls.types))
        a = np.array([tuple(row) for row in args[0]], dtype=dtype)
        return a.view(cls)


class QueryResult(Base):

    """Represents a query from the database as a numpy rec array"""

    columns = 'v u vv uu dist_v dist_u weight'.split()
    types = [np.int64, np.int64, np.int64, np.int64,
             np.float32, np.float32, np.float32]

    @property
    def v(self):
        """Get column v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'v')

    @property
    def u(self):
        """Get column u

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'u')

    @property
    def vv(self):
        """Get column vv - written v prime (v') in the paper
        where v' is a query node within
        hopping distance h of query node v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'vv')

    @property
    def uu(self):
        """Get column uu - written u prime (u') in the paper
        where u' is a target node within
        hopping distance h of target node u

        values less than zero indicate that uu (u') has
        no corresponding matches to any node v'

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'uu')

    @property
    def dist_v(self):
        """The hopping distance between query node v and query node vv (v')

        Returns:
            np.ndarray -- array of hopping distances as integers
        """

        return getattr(super(), 'dist_v')

    @property
    def dist_u(self):
        """The hopping distance between target node u and target node uu (u')

        Returns:
            np.ndarray -- array of hopping distances as integers
        """

        return getattr(super(), 'dist_u')

    @property
    def weight(self):
        """String matching score between uu (u') and vv (v')

        Returns:
            np.ndarray -- array of floating point weights
        """

        return getattr(super(), 'weight')

    def __repr__(self):
        return 'QueryResult(records={}, dtypes={})'.format(
            [record for record in self], self.types
        )


class NeighbourHoodMatchingCosts(Base):

    """Represents a table of all valid neighbourhood matching costs"""

    columns = 'v u vv uu cost'.split()
    types = [np.int64, np.int64, np.int64, np.int64, np.float32]

    def __getitem__(self, indx):
        """Get the row representing the neighbourhood matching costs at index "indx"

        Arguments:
            indx {int, slice} -- index into the table

        Returns:
            NeighbourHoodMatchingCosts -- NeighbourHoodMatchingCosts
            sliced by indx
        """
        return super().__getitem__(indx)

    @property
    def v(self):
        """Get column v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'v')

    @property
    def u(self):
        """Get column u

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'u')

    @property
    def vv(self):
        """Get column vv - written v prime (v') in the paper
        where v' is a query node within
        hopping distance h of query node v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'vv')

    @property
    def uu(self):
        """Get column uu - written u prime (u') in the paper
        where u' is a target node within
        hopping distance h of target node u

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'uu')

    @property
    def cost(self):
        """Get column cost - all valid neighbourhood matching costs.

        Eq 2 in the paper - multiplied by 1 - lambda

        Returns:
            np.ndarray -- array of costs and floats
        """

        return getattr(super(), 'cost')

    def __repr__(self):
        return 'NeighbourHoodMatchingCosts(records={}, dtypes={})'.format(
            [record for record in self], self.types
        )


class PartialMatchingCosts(Base):

    """ A table representing all valid partial matching costs """

    columns = 'v u vv cost'.split()
    types = [np.int64, np.int64, np.int64, np.float32]

    @property
    def v(self):
        """Get column v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'v')

    @property
    def u(self):
        """Get column u

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'u')

    @property
    def vv(self):
        """Get column vv - written v prime (v') in the paper
        where v' is a query node within
        hopping distance h of query node v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'uu')

    @property
    def cost(self):
        """Get column cost - all valid partial matching costs.

        Eq 13 in the paper (W) - but with beta multiplied by
        a factor of 1 - lambda

        Returns:
            np.ndarray -- array of costs as floats
        """

        return getattr(super(), 'cost')

    def __repr__(self):
        return 'PartialMatchingCosts(records={}, dtypes={})'.format(
            [record for record in self], self.types
        )


class InferenceCost(Base):

    """ A table representing all valid inference costs between query node
    u and target node v"""

    columns = 'v u cost'.split()
    types = [np.int64, np.int64, np.float32]

    @property
    def v(self):
        """Get column v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'v')

    @property
    def u(self):
        """Get column u

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'u')

    @property
    def cost(self):
        """Get column cost - all valid inference costs for
        query node v and target node u.

        Eq 14 in the paper (U)

        Returns:
            np.ndarray -- array of costs as floats
        """

        return getattr(super(), 'cost')

    def __repr__(self):
        return 'InferenceCost(records={}, dtypes={})'.format(
            [record for record in self], self.types
        )


class OptimalMatch(Base):

    """Table representing the cost of the optimal
    match for query node v going to u"""

    columns = 'v u cost'.split()
    types = [np.int64, np.int64, np.float32]

    @property
    def v(self):
        """Get column v

        Returns:
            np.ndarray -- array of query node ids as integers
        """

        return getattr(super(), 'v')

    @property
    def u(self):
        """Get column u

        Returns:
            np.ndarray -- array of target node ids as integers
        """

        return getattr(super(), 'u')

    @property
    def cost(self):
        """Get column cost - the optimal matching cost for u going to v.

        Eq 10 in the paper (O)

        Returns:
            np.ndarray -- array of costs as floats
        """

        return getattr(super(), 'cost')


# TODO: This seems a big ugly as a class and not consistent with the rest
# of the code
class Refiner:
    """Take each of the matches and recursivly find
    all of their neighbours via a greedy algorithm"""

    def __init__(
            self, neighbourhood_matching_costs: NeighbourHoodMatchingCosts):
        """Initialise a refiner using a Frame instance

        For each pair (v, u) find the lowest cost adjacent pair (vv, uu)

        Arguments:
            frame {[Frame]} -- A frame constructed
            records returned by the database
        """

        # group all of the pairs (v, u)
        matches = {tuple(k): v for k, v in zip(
            *group_by(['v', 'u'], neighbourhood_matching_costs))}

        # get the costs cost neighbouring pair (vv, uu)
        neighbours = {
            k: group_by_first(['v', 'u', 'vv'], v)[['vv', 'uu']].tolist()
            for k, v in matches.items()
        }

        self.neighbours = {
            k: [item for item in v if self.valid_neighbours(k, item)]
            for k, v in neighbours.items()
        }

    def __call__(self, seed, result):
        """Given a pair v, u,
        greedily find the lowest cost neighbours
        recursivly covering the whole graph
        without cyclic paths

        Arguments:
            seed {[int, int]} -- u, v pair

        Keyword Arguments:
            result {[int, int]} -- stores the result between
            recursive calls (default: {None})

        Returns:
            [type] -- List of query_node, target_node
            id pairs constituting a result
        """

        # To not add a new result if we've already visited this query node
        if seed[0] in set(query_node for query_node, target_node in result):
            return

        # Now do the same for all the neighbours of the new result
        result += [seed]
        for neighbour in self.neighbours[seed]:
            self(neighbour, result)
        return

    @staticmethod
    def valid_neighbours(first: tuple, second: tuple):
        """Function that governs a valid hop between nodes

        Arguments:
            first {int, int} -- source query_node, target_node id pair
            second {int, int} -- target query_node, target_node id pair

        Returns:
            Bool -- True is a valid transition
        """

        # transitions to self are not valid
        if first == second:
            return False
        # transitions to missing target nodes are not valid
        if any(u < 0 for u in second):
            return False
        return True


def _get_matching_costs(
        records: List[tuple], hopping_distance, lmbda=.3, alpha=.3, ):
    """Create a table of matching costs from a
    table of query results using equation 2
    Equivalent to the first term of equation 13

    Returns:
        NeighbourHoodMatchingCosts -- table of all valid matching costs
        QueryResult -- query result as a numpy rec
        array rather than a list of tuples
    """

    # convert NaN records into negative numbers so they can be stored as ints
    # using numpy
    query_result = QueryResult(
        [
            tuple(item if item is not None else -1 for item in tup)
            for tup in records
        ]
    )
    # label costs are weights in the databse
    query_result.weight = 1. - query_result.weight
    query_result = np.sort(query_result, order=[
                           'v', 'u', 'vv', 'uu', 'weight'])

    nan_idx = query_result.dist_u < 0
    dist_u = query_result.dist_u
    dist_u[nan_idx] = hopping_distance + 1

    # convert hopping distances into proximities (eq. 1)
    prox_v = _proximity(hopping_distance, alpha, query_result.dist_v)
    prox_u = _proximity(hopping_distance, alpha, dist_u)
    cost = _delta_plus(prox_v, prox_u)
    cost *= (1. - lmbda)

    arr = np.unique(
        np.rec.fromarrays(
            (query_result.v, query_result.vv, prox_v),
            dtype=[('v', np.int64), ('vv', np.int64), ('prox_v', float)]
        ),
        axis=0
    )

    vs, groups = group_by('v', arr)
    beta_ = {v[0]: sum(group['prox_v']) for v, group in zip(vs, groups)}
    beta = np.vectorize(lambda x: beta_[x])
    neighbourhood_matching_costs = np.rec.fromarrays([
            query_result.v,
            query_result.u,
            query_result.vv,
            query_result.uu,
            cost
        ],
        dtype=list(zip(
            NeighbourHoodMatchingCosts.columns,
            NeighbourHoodMatchingCosts.types
        ))
    ).view(NeighbourHoodMatchingCosts)
    return neighbourhood_matching_costs, query_result, beta


def _get_partial_inference_costs(
        neighbourhood_matching_costs: NeighbourHoodMatchingCosts,
        beta
) -> PartialMatchingCosts:
    """get the lowest cost neighbourhood matching cost for each partial match

    Arguments:
        neighbourhood_matching_costs {NeighbourHoodMatchingCosts} -- sorted
        neighbourhood_matching_costs

    Returns:
        PartialMatchingCosts
    """

    grouped = group_by_first(['v', 'u', 'vv'], neighbourhood_matching_costs)
    partial_matching_costs = np.rec.fromarrays([
            grouped.v, grouped.u, grouped.vv, grouped.cost
        ],
        dtype=list(zip(
            PartialMatchingCosts.columns,
            PartialMatchingCosts.types
        ))
    ).view(PartialMatchingCosts)
    partial_matching_costs.cost /= beta(partial_matching_costs.v)
    return partial_matching_costs


def _get_inference_costs(
        partial_matching_costs: PartialMatchingCosts) -> InferenceCost:
    """sum partial matching cost for each query node target node pair (v, u)

    Arguments:
        partial_matching_costs {PartialMatchingCosts} -- sorted
        partial_matching_costs

    Returns:
        InferenceCost
    """

    # TODO: inference_costs include the label weight
    keys, groups = group_by(['v', 'u'], partial_matching_costs)
    summed = starmap(
        lambda key, group: (key[0], key[1], np.sum(group.cost) / len(group)),
        zip(keys, groups)
    )
    return InferenceCost(summed)


def _get_optimal_match(inference_costs: InferenceCost) -> OptimalMatch:
    """Get the lowest cost match for each query node v

    Arguments:
        inference_costs {InferenceCost} -- sorted inference costs

    Returns:
        OptimalMatch
    """

    return group_by_first('v', inference_costs)


def solve(records: List[tuple], max_iters=10, hopping_distance=2):
    """Generate a set of subgraph matches and costs from a query result

    Arguments:
        records {List[tuple]}

    """

    # initialisation
    finished, iters = False, 0
    prv_optimum_match = None

    neighbourhood_matching_costs, query_result, beta = _get_matching_costs(
        records, hopping_distance)
    # keep a copy for successive iterations
    neighbourhood_matching_costs_cpy = neighbourhood_matching_costs.copy()

    label_costs = {
        (record.v, record.u): record.weight
        for record in query_result
    }
    label_costs_func = np.vectorize(lambda x: label_costs.get(tuple(x)))

    while True:

        # first optimisation
        neighbourhood_matching_costs = np.sort(
            neighbourhood_matching_costs,
            order=['v', 'u', 'vv', 'cost'],
            axis=0
        )
        partial_inference_costs = _get_partial_inference_costs(
            neighbourhood_matching_costs, beta)
        inference_costs = _get_inference_costs(partial_inference_costs)
        additional_costs = np.array([label_costs.get(tuple(x)) for x in inference_costs[['v', 'u']]], dtype=np.float32)
        inference_costs.cost += additional_costs

        # second optimisation
        inference_costs = np.sort(inference_costs, order=['v', 'cost'])
        optimum_match = _get_optimal_match(inference_costs)
        inference_costs_dict = {
            (record.v, record.u): record.cost for record in inference_costs}
        apply = np.vectorize(
            lambda x: inference_costs_dict.get(tuple(x), iters))
        neighbourhood_matching_costs = neighbourhood_matching_costs_cpy.copy()
        neighbourhood_matching_costs.cost += apply(
            neighbourhood_matching_costs[['vv', 'uu']])
        iters += 1

        # test for convergance
        if prv_optimum_match is not None:
            diff = prv_optimum_match[['v', 'u']] == optimum_match[['v', 'u']]
            finished = (sum(diff) / len(optimum_match)) > .9
        if finished:
            break

        prv_optimum_match = optimum_match

        if iters >= max_iters:
            break

    neighbourhood_matching_costs = np.sort(
        neighbourhood_matching_costs,
        order=['v', 'u', 'vv', 'cost'],
        axis=0
    )
    refine = Refiner(neighbourhood_matching_costs)

    # normalise inference costs by the number of iterations
    inference_costs.cost /= iters
    inference_costs_dict = {
        (record.v, record.u): record.cost for record in inference_costs}

    inference_costs = np.sort(inference_costs, order=['cost'])
    subgraph_matches = []
    for seed in inference_costs[['v', 'u']]:
        subgraph_match = []
        refine(tuple(seed), subgraph_match)
        subgraph_match = sorted(subgraph_match)
        if subgraph_match not in subgraph_matches:
            subgraph_matches.append(subgraph_match)
    target_edges = group_by_first(['u', 'uu', 'dist_u'], query_result)[
        ['u', 'uu', 'dist_u']]
    mask = target_edges['dist_u'] > 0
    mask *= target_edges['dist_u'] <= 1
    mask *= target_edges['u'] < target_edges['uu']
    mask > 0
    target_edges = np.sort(target_edges[mask])
    return inference_costs_dict, subgraph_matches, iters, len(
        optimum_match), target_edges
